list(
  # Adjustment name ----
  Name = "Reporting Delays",

  # Input parameters to the adjustment function ----
  Parameters = list(
    # Parameter 1: a specification with label, type and default value
    startYear = list(
      label = "Diagnosis start year",
      value = 2000L,
      input = "numeric"),
    # Parameter 2: a specification with label, type and default value
    endQrt = list(
      label = "Notification end quarter",
      value = 2017.25,
      input = "numeric"),
    stratGender = list(
      name = "stratGender",
      label = "Gender",
      value = FALSE,
      input = "checkbox"),
    stratTrans = list(
      name = "stratTrans",
      label = "Transmission",
      value = FALSE,
      input = "checkbox"),
    stratMigr = list(
      name = "stratMigr",
      label = "Migration",
      value = FALSE,
      input = "checkbox")
  ),

  # Names of packages that must be made available to the adjustment function ----
  RequiredPackageNames = c(),

  ## Adjustment function ----
  AdjustmentFunction = function(inputData, parameters) {

    # Work on a copy
    outputData <- copy(inputData)

    # Color palette for the plots
    # First color: background of confidence interval for the estimated count
    # Second color: estimated count
    # Third color: reported count
    colorPalette <- c("#AAAAAA", "#E69F00", "#56B4E9")

    # Separator used for creating a composite of stratum columns. Should not occur in the stratum values.
    stratSep <- "_"

    # Start year
    startYear <- parameters$startYear
    # End quarter
    endQrt <- parameters$endQrt
    # Stratifiation columns
    stratVarNames <- c()
    if (parameters$stratGender) {
      stratVarNames <- union(stratVarNames, "Gender")
    }
    if (parameters$stratTrans) {
      stratVarNames <- union(stratVarNames, "Transmission")
    }
    if (parameters$stratMigr) {
      stratVarNames <- union(stratVarNames, "Migr")
    }
    stratVarNames <- stratVarNames[stratVarNames %in% colnames(inputData)]

    # B) PROCESS DATA ---------------------------------------------------------------------------------

    # Add dummy "Imputation" column if not found
    hasImputationCol <- "Imputation" %in% colnames(inputData)
    if (!hasImputationCol) {
      outputData[, Imputation := 1L]
    }
    # Number of imputations in the data (at least one - dummy)
    nimp <- outputData[, max(1, length(unique(Imputation)) - 1)]

    # Make sure the strata columns exist in the data
    stratVarNames <- stratVarNames[stratVarNames %in% colnames(inputData)]
    stratVarNamesImp <- union("Imputation", stratVarNames)

    # Create intermediate variables
    outputData[, ":="(
      NotificationTime = DateOfNotificationYear + 1/4 * DateOfNotificationQuarter,
      DiagnosisTime = DateOfDiagnosisYear + 1/4 * DateOfDiagnosisQuarter
    )]

    # Filter
    outputData <- outputData[!is.na(DiagnosisTime) & !is.na(NotificationTime)]
    outputData <- outputData[DiagnosisTime >= max(startYear, min(NotificationTime, na.rm = TRUE))
                             & NotificationTime <= endQrt]
    outputData[, VarX := 4 * (NotificationTime - DiagnosisTime)]
    outputData[VarX < 0, VarX := NA]
    outputData[, ":="(
      MinNotificationTime = min(NotificationTime, na.rm = TRUE),
      MaxNotificationTime = max(NotificationTime, na.rm = TRUE)
    ), by = .(ReportingCountry)]

    outputData[, ":="(
      VarT = 4 * (pmin(MaxNotificationTime, endQrt) - DiagnosisTime) + 1,
      Tf = 4 * (pmin(MaxNotificationTime, endQrt) - pmax(MinNotificationTime, startYear)),
      ReportingDelay = 1L
    )]
    outputData[, ":="(
      VarXs = Tf - VarX,
      VarTs = Tf - VarT
    )]
    # NOTE: Otherwise survival model complains
    outputData <- outputData[VarXs > VarTs]

    if (nrow(outputData) > 0) {

      # Create a stratum variable - will be used to merge with the estimations,
      # takes missing categories as separate categories
      outputData[, Stratum := survival::strata(.SD, shortlabel = TRUE, sep = stratSep, na.group = TRUE),
                 .SDcols = stratVarNamesImp]

      # Fit a stratified survival model
      model <- outputData[, survival::Surv(time = VarTs,
                                           time2 = VarXs,
                                           event = ReportingDelay)]
      fit <- outputData[, survival::survfit(model ~ Stratum)]
      if (is.null(fit$strata)) {
        strata <- 1L
      } else {
        strata <- fit$strata
      }

      # Aggregate and keep only required dimensions
      agregat <- outputData[, .(Count = .N),
                            by = .(DiagnosisTime,
                                   YearOfDiagnosis = DateOfDiagnosisYear,
                                   ReportingCountry,
                                   VarT,
                                   Stratum)]
      # Recreating stratum variables to assign them to the delay distribution dataset
      fitStratum <- data.table(
        Delay = fit$time,
        P = fit$surv,
        Var = fit$std.err^2,
        Stratum = factor(rep(seq_along(strata), strata),
                         labels = levels(outputData$Stratum)))
      fitStratum[, (stratVarNamesImp) := tstrsplit(Stratum, stratSep)]
      fitStratum[, VarT := max(Delay) - Delay]
      fitStratum <- fitStratum[VarT >= 0]
      # Merge fit data
      agregat <- fitStratum[agregat,
                            on = .(VarT, Stratum)]

      # Compute estimated count and its variance
      agregat[, ":="(
        EstCount = Count / P,
        EstCountVar = (Count * (Count + 1) / P^4 * Var) + Count * (1 - P) / P^2
      )]

      # c) TOTAL PLOT -----------------------------------------------------------------------------------

      # Aggregating by year
      agregatByYear <- agregat[, .(Count = sum(Count),
                                   EstCount = sum(EstCount),
                                   EstCountVar = sum(EstCountVar)),
                               by = .(Imputation, YearOfDiagnosis)]

      # Combine estimates obtained for each imputation
      agregatByYear <- agregatByYear[, .(Count = mean(Count),
                                         EstCount = mean(EstCount),
                                         EstCountVarWi = mean(EstCountVar),
                                         EstCountVarBe = var(EstCount)),
                                     by = .(YearOfDiagnosis)]
      agregatByYear[is.na(EstCountVarBe), EstCountVarBe := 0]

      agregatByYear[, ":="(
        EstCountVar = EstCountVarWi + (1 + 1/nimp) * EstCountVarBe,
        EstCountVarWi = NULL,
        EstCountVarBe = NULL
      )]

      agregatByYear[, ":="(
        LowerEstCount = EstCount - 1.96 * sqrt(EstCountVar),
        UpperEstCount = EstCount + 1.96 * sqrt(EstCountVar)
      )]

      agregatPlotYr <- ggplot(agregatByYear, aes(x = YearOfDiagnosis)) +
        geom_ribbon(aes(ymin = pmin(pmax(LowerEstCount, 0), max(EstCount)),
                        ymax = pmin(pmax(UpperEstCount, 0), max(EstCount)),
                        fill = "95% confidence interval\nfor estimated total count"),
                    alpha = 0.4) +
        scale_fill_manual("Bounds", values = colorPalette[1]) +
        geom_line(aes(y = Count, color = "Reported"), size = 0.2) +
        geom_line(aes(y = EstCount, color = "Estimated total"), size = 0.5) +
        scale_colour_manual("Counts", values = colorPalette[2:3]) +
        ggtitle("Reported and estimated total count of cases") +
        xlab("Diagnosis year") +
        ylab("Count of HIV cases")

      # D) STRATIFIED PLOT (OPTIONAL) -------------------------------------------------------------------
      plotList <- NULL
      if (length(stratVarNames) > 0) {
        # Stratification
        colNames <- union(c("YearOfDiagnosis", "Count", "EstCount", "EstCountVar"),
                          stratVarNamesImp)
        # Keep only required columns, convert data to "long" format...
        agregatByYearStrat <- melt(agregat[, ..colNames],
                                   measure.vars = stratVarNames,
                                   variable.name = "Stratum",
                                   value.name = "StratumValue")
        # ...and aggregate
        agregatByYearStrat <- agregatByYearStrat[, .(Count = sum(Count),
                                                     EstCount = sum(EstCount),
                                                     EstCountVar = sum(EstCountVar)),
                                                 by = .(Imputation, YearOfDiagnosis, Stratum, StratumValue)]

        # Combine estimates obtained for each imputation
        agregatByYearStrat <- agregatByYearStrat[, .(Count = mean(Count),
                                                     EstCount = mean(EstCount),
                                                     EstCountVarWi = mean(EstCountVar),
                                                     EstCountVarBe = var(EstCount)),
                                                 by = .(YearOfDiagnosis, Stratum, StratumValue)]
        agregatByYearStrat[is.na(EstCountVarBe), EstCountVarBe := 0]
        agregatByYearStrat[, ":="(
          EstCountVar = EstCountVarWi + (1 + 1/nimp) * EstCountVarBe,
          EstCountVarWi = NULL,
          EstCountVarBe = NULL
        )]

        agregatByYearStrat[, ":="(
          LowerEstCount = EstCount - 1.96 * sqrt(EstCountVar),
          UpperEstCount = EstCount + 1.96 * sqrt(EstCountVar),
          EstCountVar = NULL
        )]

        plotList <- lapply(stratVarNames, function(stratum) {
          ggplot(agregatByYearStrat[Stratum == stratum], aes(x = YearOfDiagnosis)) +
            geom_ribbon(aes(ymin = pmin(pmax(LowerEstCount, 0), max(EstCount)),
                            ymax = pmin(pmax(UpperEstCount, 0), max(EstCount)),
                            fill = "95% confidence interval\nfor estimated total count"),
                        alpha = 0.4) +
            scale_fill_manual("Bounds", values = colorPalette[1]) +
            geom_line(aes(y = Count, color = "Reported"), size = 0.2) +
            geom_line(aes(y = EstCount, color = "Estimated total"), size = 0.5) +
            scale_colour_manual("Counts", values = colorPalette[2:3]) +
            facet_wrap(~ StratumValue, ncol = 2) +
            ggtitle(paste("Reported and estimated total count of cases, by", stratum)) +
            xlab("Diagnosis year") +
            ylab("Count of HIV cases")
        })
      }
    } else {
      agregatPlotYr <- NULL
      plotList <- NULL
      agregat <- NULL
    }

    artifacts <- list(OutputPlotYr = agregatPlotYr,
                      OutputPlotStrat = plotList)

    return(list(Table = agregat,
                Artifacts = artifacts))
  }
)
